/* Copyright 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Implements a data cache
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */
#include "thinc_cache.h"
#include <stdlib.h>

#define T_ARBITRARY	481124

static void thinc_cache_delete_el(thinc_cache *, thinc_cache_el *);
static thinc_cache_el* thinc_cache_get_el(thinc_cache *, unsigned long int );

#define thinc_zero_el(el, el_priv) \
	(el)->hash = 0; \
	(el)->flags = T_EMPTY; \
	(el)->priv = el_priv;

#define thinc_cache_idx(cache, h) \
	((h) & thinc_hashmask((cache)->idx_len));

#define thinc_do_hash(cache, h, idx) \
	h = thinc_hash(data, length, T_ARBITRARY); \
	idx = thinc_cache_idx(cache, h);

thinc_cache *thinc_cache_init(int idx_len)
{
	thinc_cache *cache;
	register unsigned int i;

	cache = malloc(sizeof(thinc_cache));
	if(!cache)
		return NULL;

	cache->idx_len = idx_len;
	cache->total = thinc_cache_size(cache);
	cache->active = cache->collisions = 0;

	cache->cache = malloc(cache->total* sizeof(thinc_cache_el));
	if(!cache->cache) {
		free(cache);
		return NULL;
	}

	for(i=0; i<cache->total; ++i) {
		thinc_zero_el(&cache->cache[i], NULL);
	}

	return cache;
}

int thinc_cache_get_size(thinc_cache *cache)
{
	return cache->total;
}

int thinc_cache_get_active(thinc_cache *cache)
{
	return cache->active;
}

int thinc_cache_get_collisions(thinc_cache *cache)
{
	return cache->collisions;
}

void thinc_cache_destroy(thinc_cache* cache)
{
	if(!cache)
		return;
	thinc_cache_empty(cache);
	free(cache->cache);
	free(cache);
}

void thinc_cache_destroy_cb(thinc_cache* cache, 
	void (*callback)(void *, void *), void *ptr)
{
	thinc_cache_empty_cb(cache, callback, ptr);
	free(cache->cache);
	free(cache);
}

void thinc_cache_empty(thinc_cache* cache)
{
	register unsigned int i;

	for(i=0; i<cache->total; ++i) {
		free(cache->cache[i].priv);
		thinc_zero_el(&cache->cache[i], NULL);
	}

	cache->active = 0;
}

void thinc_cache_empty_cb(thinc_cache* cache, void (*callback)(void *, void *),
	void *ptr)
{
	register unsigned int i;

	for(i=0; i<cache->total; ++i) {
		callback(ptr, cache->cache[i].priv);
		thinc_zero_el(&cache->cache[i], NULL);
	}

	cache->active = 0;
}

/* Return T_YES if h is in cache. T_NO if not */
int thinc_cache_query_id(thinc_cache *cache, unsigned long int h)
{
	thinc_cache_el *e;

	e = thinc_cache_get_el(cache, h);
	return (e ? T_YES : T_NO);
}

/* Return T_YES if data is in cache. T_NO if not. The hash is returned
 * in hash */
int thinc_cache_query_data(thinc_cache *cache, register char *data, 
	register unsigned long int length, unsigned long int *hash)
{
	unsigned int idx;
	unsigned long int h;

	thinc_do_hash(cache, h, idx);
	if(hash)
		*hash = h;
	return thinc_cache_query_id(cache, h);
}

unsigned long int thinc_cache_get_hash(thinc_cache *cache, 
	register char *data, register unsigned long int length)
{
	return thinc_hash(data, length, T_ARBITRARY);
}

/* returns hash of added element. If element is already in cache, it
 * just returns the hash */
unsigned long int thinc_cache_add(thinc_cache *cache, 
	register char *data, register unsigned long int length, 
	void *priv)
{
	unsigned long int h;
	int ret;

	ret = thinc_cache_query_data(cache, data, length, &h);
	if(ret == T_YES) 
		return h;

	thinc_cache_add_hash(cache, h, priv);
	return h;
}

int thinc_cache_add_hash(thinc_cache *cache, unsigned long int hash,
	void *priv)
{
	unsigned int idx, curr, mask;
	thinc_cache_el *e;

	idx = thinc_cache_idx(cache, hash);
	e = &cache->cache[idx];
	if(e->flags & T_ACTIVE) {
		curr = idx;
		mask = thinc_hashmask((cache)->idx_len);
		do {
			curr = (curr+1)&mask;
			e = &cache->cache[curr];
		} while((e->flags & T_ACTIVE) && (curr != idx));

		if(curr == idx) {
			/* no hay cama pa tanta gente */
			cache->collisions++;
			thinc_cache_delete_el(cache, e);
		}
	}

	e->hash = hash;
	e->flags = T_ACTIVE;
	e->priv = priv;
	cache->active++;

	return T_OK;
}

static void thinc_cache_delete_el(thinc_cache *cache, thinc_cache_el *e)
{
	free(e->priv);
	thinc_zero_el(e, NULL);
	if(cache->active)
		cache->active--;
}

void thinc_cache_delete(thinc_cache *cache, unsigned long int hash)
{
	thinc_cache_el *e;
	unsigned int idx;

	idx = thinc_cache_idx(cache, hash);
	e = &cache->cache[idx];

	thinc_cache_delete_el(cache, e);
}

static thinc_cache_el* thinc_cache_get_el(thinc_cache *cache, 
	unsigned long int h)
{
	unsigned int idx;
	register unsigned int curr, mask, gotit;
	register thinc_cache_el* e;

	idx = (h & thinc_hashmask(cache->idx_len));
	e = &cache->cache[idx];
	if(e->flags & T_ACTIVE) {
		if(e->hash == h)
			return e;
		else {
			/* may be later on in the array */
			mask = thinc_hashmask((cache)->idx_len);
			gotit = T_NO;
			curr = (idx+1)&mask;
			e = &cache->cache[curr];
			while((e->flags & T_ACTIVE) && (curr != idx)) {
				if(e->hash == h) {
					gotit = T_YES;
					break;
				}
				curr = (curr+1)&mask;
				e = &cache->cache[curr];
			}
			return (gotit ? e : NULL);
		}
	} else return NULL;
}

/* returns priv of element. NULL if element is not in cache */
void* thinc_cache_get(thinc_cache *cache, unsigned long int hash)
{
	thinc_cache_el *e;
	e = thinc_cache_get_el(cache, hash);
	if(!e)
		return NULL;
	else
		return e->priv;
}
