/* Copyright 2005, 2007 Ricardo Baratto */
/* Copyright 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/* 
 * thinc_zsound_dec.c: routines for decoding sound
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#include "thinc_zsound.h"
#include "thincCommon.h"
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#define __TZ_IN_LIB
#include "thinc_zsound_priv.h"

/* sound decoding library specific structs (careful, names are the same
 * from encoder) */

/* for vorbis we use vorbisfile instead of libvorbis (which is quite a mess) */
struct tz_vorbis {
	OggVorbis_File vf;
};

struct tz_speex {
		/* nothing for now as we have no speex support */
};

tz_funcs_dec(vorbis);
tz_funcs_dec(speex);

#define tz_fill_funcs(zs, name)			\
	zs->init = tz_init_##name;		\
	zs->end = tz_end_##name;		\
	zs->prepare = tz_prepare_##name;	\
	zs->decode = tz_decode_##name;	\
	zs->finish = tz_finish_##name;

/* local callback functions for vorbis */
static size_t tz_vorbis_cb_read(void *ptr, size_t size, size_t nmemb, 
	void *datasource);
static int tz_vorbis_cb_close (void *datasource);
static int tz_vorbis_cb_seek(void *datasource, ogg_int64_t offset, int whence);

/* real code starts here */
struct thinc_zsound* thinc_zsound_dec_init(int which, ...)
{
	va_list params;
	struct thinc_zsound *zs;
	int err;

	zs = malloc(sizeof(*zs));
	if(!zs)
		return NULL;

	switch(which) {
		case T_ZSOUND_VORBIS:
			tz_fill_funcs(zs, vorbis);
			break;
		case T_ZSOUND_SPEEX:
			tz_fill_funcs(zs, speex);
			break;
		default:
			free(zs);
			return NULL;
	}

	va_start(params, which);
	err = zs->init(zs, params);
	va_end(params);
	if(err == T_ERROR) {
		free(zs);
		return NULL;
	}
	zs->read = NULL;
	zs->write = NULL;
	zs->cb_handle = NULL;

	return zs;
}

/* call this after init and before doing ANYTHING else */
void thinc_zsound_dec_setcb(struct thinc_zsound *zs, 
	int (*read)(void *, void *, unsigned int), 
	void *cb_handle)
{
	TZ_DEBUG("read %p, handle %p\n", read, cb_handle);
	zs->read = read;
	zs->cb_handle = cb_handle;
}

void thinc_zsound_dec_end(struct thinc_zsound* zs)
{
	zs->end(zs);
	free(zs);
}

int thinc_zsound_dec_prepare(struct thinc_zsound *zs)
{
	return zs->prepare(zs);
}
int thinc_zsound_dec_decode(struct thinc_zsound* zs, char *decoded, int len)
{
	return zs->decode(zs, decoded, len);
}

/* done decodeing, flush all the data */
int thinc_zsound_dec_finish(struct thinc_zsound* zs)
{
	return zs->finish(zs);
}

/**** vorbis funcs ****/
/* most of this code taken from vorbisfile_example.c from libvorbis dist.
 * thanks guys! */
/* only need to call once */
static int tz_init_vorbis(struct thinc_zsound *zs, va_list params)
{
	struct tz_vorbis *vorbis;

	vorbis = malloc(sizeof(*vorbis));
	if(!vorbis)
		return T_ERROR;

	zs->handle = (void*) vorbis;
	
	return T_OK;
}

/* you should've set up the callbacks before calling this function. it will
 * read in some data using the read callback to set things up so be prepared 
 * for that (ie don't call us if you know there's no data to be read) */
static int tz_prepare_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;
	ov_callbacks cb;

	cb.read_func = tz_vorbis_cb_read;
	cb.seek_func = tz_vorbis_cb_seek;
	cb.close_func = tz_vorbis_cb_close;
	cb.tell_func = NULL;

	if(ov_open_callbacks(zs, &vorbis->vf, NULL, 0, cb) != 0) {
		TZ_DEBUG("cannot open vorbis source using callbacks\n");
		return T_ERROR;
	}

	return T_OK;
}

/* decode some data and put it in buffer. it won't decode more than len
 * bytes. returns how much data was actually decoded. if T_ZSOUND_DONE
 * is returned, we're at the end of the stream and you should call
 * finish() */
static int tz_decode_vorbis(struct thinc_zsound *zs, char* decoded, int len)
{
	struct tz_vorbis *vorbis = zs->handle;
	long err;
	int current_section; /* unused?? */

	/* FIXME: Can we have these params fixed like this??
	 * 0: Specifies big or little endian byte packing. 0 for little 
	 * endian, 1 for b ig endian. Typical value is 0.
	 * 2: specifies word size. Possible arguments are 1 for 8-bit samples, 
	 * or 2 or 16-bit samples. Typical value is 2.
	 * 1: Signed or unsigned data. 0 for unsigned, 1 for signed. 
	 * Typically 1.
	 */
	err = ov_read(&vorbis->vf, decoded, len, 0, 2, 1, &current_section);
	if(err == 0) {
		/* done */
		return T_ZSOUND_DONE;
	} else if(err < 0) {
		/* error in the stream.  May or may not be a problem, depends
		 * on app. in UDP this may happen and we probably just want to
		 * ignore it, so doing that for now ... */
	}
	return err;
}

static int tz_finish_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;

	ov_clear(&vorbis->vf);
	return T_OK;
}

static void tz_end_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;
	free(vorbis);
	zs->handle = NULL;
}

/* some callbacks needed by vorbisfile but which we have no need for and
 * there's no point in asking users to provide */
/* from docs: The seek function must always return -1 (failure) if the given
 * data abstraction is not seekable. It may choose to always return -1 if the
 * application desires libvorbisfile to treat the Vorbis data strictly as a
 * stream (which makes for a less expensive open operation).  */
static int tz_vorbis_cb_seek(void *datasource, ogg_int64_t offset, int whence)
{
	return -1;
}

/* The close function is assumed to succeed. */
static int tz_vorbis_cb_close (void *datasource)
{
	return 0;
}

/* our read function closes the gap between the vorbisfile abstraction and
 * the read we require from callers */
static size_t tz_vorbis_cb_read(void *ptr, size_t size, size_t nmemb, 
	void *datasource)
{
	struct thinc_zsound* zs = datasource;

	return zs->read(zs->cb_handle, ptr, size*nmemb);
}

/**** speex funcs ****/
/* TODO */
static int tz_init_speex(struct thinc_zsound *zs, va_list params)
{
	return T_ERROR;
}

static int tz_prepare_speex(struct thinc_zsound *zs)
{
	return T_ERROR;
}

static int tz_decode_speex(struct thinc_zsound *zs, char *buffer, int len)
{
	return T_ERROR;
}

static int tz_finish_speex(struct thinc_zsound *zs)
{
	return T_ERROR;
}

static void tz_end_speex(struct thinc_zsound *zs)
{
	free(zs->handle);
}
