/* Copyright 2004, 2005, 2007 Ricardo Baratto */
/* Copyright 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * resize functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* ***
 * fant resizing: from page 156 of Digital Image Warping by George Woldberg. 
 * Adapted to resize 3 color channels at the same time, and to find pixels in
 * columns
 * ***
 */

/* TODO: Can we do rows and columns in one pass? (no need for tmp) */

/* set tabstop 8 */

#include "thinc_resize.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define T_RZ_TESTING

/* FIXME:
 * Doesn't work as it is. Only precomputes correctly for some scales
#define PRECOMP
 */

#ifdef T_RZ_TESTING
# include <assert.h>
#endif

/* clips value v to range [low,high] */
#define CLIP(v, low, high) ((v) < (low) ? (low) : ((v) > (high) ? (high):(v)))
#define CLIP_UP(v, high) ((v) > (high) ? (high):(v))

#define MAX_GRAY	0xff
#define MAX_2BYTES	0xffff

#define DOUBLE_BASE	128.0
#define DOUBLE_ROUND	0.5

#define FP_SHIFT	7U	
#define FP_ROUND	64
#define FP_ONE		(1<<FP_SHIFT)

#define fp_in(d) \
	( (unsigned int) ( ( (d)*DOUBLE_BASE) + DOUBLE_ROUND) )

#define fp_out(fp) \
	( ( (fp) + FP_ROUND ) >> FP_SHIFT)

#define fp_mult(a, b) \
	( ( (a) * (b) ) >> FP_SHIFT)

/** resize using fant's resampling algorithm using fixed point arithmetic ***/

/* TODO: Make yuv use fixed point functions */
static int thinc_fant_simple_row8(char *in, char *out,
	unsigned int ineln, unsigned int outlen);
/* TODO: Make yuv use fixed point functions */
static int thinc_fant_simple_col8(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride);

static int thinc_fant_simple_row8_fp(char *in, char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_fant_simple_col8_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_fant_simple_row16_fp(char *in, char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_fant_simple_col16_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);
static int thinc_resize_yv12_help(char *in, char *out,
	unsigned int inw, unsigned int inh, unsigned int outw, 
	unsigned int outh, unsigned int ipitch, unsigned int opitch,
	unsigned char first);

static int thinc_fant_simple_row32_fp(char *in, char *out,
	unsigned int ineln, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale);

static int thinc_fant_simple_col32_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale);

int thinc_resize8(char *in, char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	register unsigned int i;
	register char *inc, *outc;
	char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh;
	outlen = outw*outh;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row8_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		char *from, *to;
		from = tmp;
		to = out;
		for(i=0; i<outh; ++i) {
			memcpy(to, from, outw);
			from += tmpstride;
			to += outstride;
		}
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col8_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			++inc;	/* 8 bits */
			++outc;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}

int thinc_resize16(char *in, char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 2;
	register unsigned int i;
	register char *inc, *outc;
	char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row16_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		char *from, *to;
		unsigned int len = outw*bpp;
		from = tmp;
		to = out;
		for(i=0; i<outh; ++i) {
			memcpy(to, from, len);
			from += tmpstride;
			to += outstride;
		}
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col16_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			inc += 2;	/* 32 bits */
			outc += 2;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}

int thinc_resize32(char *in, char *out, unsigned int inw, 
	unsigned int inh, unsigned int outw, unsigned outh, 
	unsigned int instride, unsigned int outstride)
{
	const int bpp = 4;
	register unsigned int i;
	register char *inc, *outc;
	char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = inw*inh*bpp;
	outlen = outw*outh*bpp;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		/* tmp is inh*outw */
		tmpstride = outw*bpp;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row32_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		char *from, *to;
		unsigned int len = outw*bpp;
		from = tmp;
		to = out;
		for(i=0; i<outh; ++i) {
			memcpy(to, from, len);
			from += tmpstride;
			to += outstride;
		}
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col32_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			inc += 4;	/* 32 bits */
			outc += 4;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}

int thinc_resize_yv12(
	char *iny, char *inu, char *inv,
	char *outy, char *outu,  char *outv,
	unsigned int inw, unsigned int inh, unsigned int outw, unsigned outh,
	unsigned int yipitch, unsigned int uvipitch,
	unsigned int yopitch, unsigned int uvopitch)
{
	unsigned int yinw, yinh, youtw, youth;
	unsigned int uvinw, uvinh, uvoutw, uvouth;
	int err;

	yinw = inw, yinh = inh;
	uvinw = inw >> 1;
	uvinh = inh >> 1;
	youtw = outw, youth = outh;
	uvoutw = outw >> 1;
	uvouth = outh >> 1;

	/* one plane at a time */
	err = thinc_resize_yv12_help(iny, outy, yinw, yinh, youtw, youth,
		yipitch, yopitch, T_TRUE);
	if(err == T_ERROR)
		return T_ERROR;
	err = thinc_resize_yv12_help(inu, outu, uvinw, uvinh, uvoutw, uvouth, 
		uvipitch, uvopitch, T_FALSE);
	if(err == T_ERROR)
		return T_ERROR;
	err = thinc_resize_yv12_help(inv, outv, uvinw, uvinh, uvoutw, uvouth, 
		uvipitch, uvopitch, T_FALSE);
	if(err == T_ERROR)
		return T_ERROR;

	return T_OK;
}

/* the idea is to call this function sequentially for each of the three
 * channels. The first time (y) pass T_TRUE so that temporary memory
 * is allocated. Since the y channel is the largest one it's ok to use
 * this memory to resize the u & v channels */
static int thinc_resize_yv12_help(char *in, char *out,
	unsigned int inw, unsigned int inh, unsigned int outw, 
	unsigned int outh, unsigned int ipitch, unsigned int opitch,
	unsigned char first)
{
	static char *tmp = NULL;
	register char *inc, *outc;
	register unsigned int i;

	if(inw == inh) {
		tmp = in;
	} else {
		if(first) {
			if(tmp)
				free(tmp);
			tmp = malloc(outw*inh);
			if(!tmp)
				return T_ERROR;
		}
		inc = in;
		outc = tmp;
		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row8(inc, outc, inw, outw);
			inc += ipitch;
			outc += outw;
		}
	}

	if(outw == outh) {
		/* TODO: Why not make out be tmp right away??? */
		char *from, *to;
		from = tmp;
		to = out;
		for(i=0; i<outh; ++i) {
			memcpy(to, from, outw);
			from += outw;
			to += opitch;
		}
	} else {
		inc = tmp;
		outc = out;
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col8(inc, outc, inh, outh,
				outw, opitch);
			++inc;
			++outc;
		}
	}

	return T_OK;
}

/**** the actual resize functions ******/
static int thinc_fant_simple_row8(char *in, char *out,
	unsigned int inlen, unsigned int outlen)
{
	unsigned int u,x;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double intensity;
	double acc;

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	acc = 0.0;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
		}
		/* use linear interpolation for reconstruction */
		intensity = (inleft * in[u]) + ((1-inleft)*in[u+1]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			acc += (intensity *inleft); /*weigthed contribution*/
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
			acc += (intensity * outleft);
			out[x] = acc * scale;
			acc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
		}
	}

	return T_OK;
}

/* do it column by column (doesn't know how to go to next column), and 
 * inlen and outlen should be the height in pixels, not bytes. stride 
 * should be in bytes */
static int thinc_fant_simple_col8(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride)
{
	unsigned int u,x, ui, xi;
	double 	scale, invscale, inleft,
		outleft; /* # input pixels that map onto 1 output pixel */ 
	double intensity;
	double acc;

	scale = (double)outlen/(double)inlen;
	invscale = 1.0/scale;
	outleft = invscale;
	inleft = 1.0;	/* entire input pixel is available */
	acc = 0.0;

	ui = 0;
	xi = 0;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ui -= instride;
		}
		/* use linear interpolation for reconstruction */
		intensity = (inleft * in[ui]) + ((1-inleft)*in[ui+instride]);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			acc += (intensity *inleft); /*weigthed contribution*/
			outleft -= inleft; /*inleft portion has been filled*/
			inleft = 1.0;
			++u;
			ui += instride;
		} else {
		/* inleft >= outleft: input pixel not entirely consumed
		 * before output pixel */
			acc += (intensity * outleft);
			out[xi] = acc * scale;
			acc = 0.0;
			inleft -= outleft;
			outleft = invscale;
			++x;
			xi += outstride;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_row8_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity;
	unsigned int acc;

	register char *inp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc = 0;

	inp = in;
	outp = out;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
			--inp;
		}
		/* use linear interpolation for reconstruction */
		intensity = inleft * inp[0] + (FP_ONE-inleft) * inp[1];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			acc += fp_mult(intensity, inleft); 

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
			++inp;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			acc += fp_mult(intensity, outleft); 

			outp[0] = CLIP_UP(
				fp_out(fp_mult(acc, fpscale)), MAX_GRAY);
			acc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp++;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col8_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int intensity;
	unsigned int acc;

	char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	acc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		intensity = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			acc += fp_mult(intensity, inleft); 
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			acc += fp_mult(intensity, outleft); 

			outp[0] = CLIP_UP(
					fp_out(fp_mult(acc, fpscale)), 
					MAX_GRAY);

			acc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}

/* 16bpp. assuming 5 reds, 6 greens, 5 blues */
#define red_mask	0xF800
#define red_shift	11
#define green_mask	0x07E0
#define green_shift	5
#define blue_mask	0x001F
#define blue_shift	0

#define get_red(pix)	( ((pix) & red_mask) >> red_shift)
#define get_green(pix)	( ((pix) & green_mask) >> green_shift)
#define get_blue(pix)	( ((pix) & blue_mask) >> blue_shift)

#define put_red(pix, red)	( (pix) |= (red_mask & ((red) << red_shift)) )
#define put_green(pix, green)   \
	        ( (pix) |= (green_mask & ((green) << green_shift)) )
#define put_blue(pix, blue)     \
	        ( (pix) |= (blue_mask & ((blue) << blue_shift)) )

static int thinc_fant_simple_row16_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;
	unsigned char r0, r1, g0, g1, b0, b1;

	char *inp, *outp;
	unsigned short sin, sinp, sout;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	outp = out;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			inp -= 2;
		}
		/* use linear interpolation for reconstruction */
		/* FIXME: This will probably break as byte ordering changes */
		/* TODO: Don't use memcpy (all we need is
		 * assign and shift? */
		memcpy(&sin, inp, sizeof(sin));
		memcpy(&sinp, &inp[2], sizeof(sinp));

		r0 = get_red(sin);
		r1 = get_red(sinp);
		g0 = get_green(sin);
		g1 = get_green(sinp);
		b0 = get_blue(sin);
		b1 = get_blue(sinp);

		rintensity = (inleft * r0) + ((FP_ONE-inleft) * r1);
		gintensity = (inleft * g0) + ((FP_ONE-inleft) * g1);
		bintensity = (inleft * b0) + ((FP_ONE-inleft) * b1);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp += 2;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			/* FIXME: This will probably break as byte ordering 
			 * changes */
			sout = 0;
			put_red(sout, CLIP_UP(
					fp_out(fp_mult(racc, fpscale)), 
					MAX_GRAY));
			put_green(sout, CLIP_UP(
					fp_out(fp_mult(gacc, fpscale)), 
					MAX_GRAY));
			put_blue(sout, CLIP_UP(
					fp_out(fp_mult(bacc, fpscale)), 
					MAX_GRAY));
			/* TODO: Don't use memcpy (all we need is
			 * assign and shift?) */
			memcpy(outp, &sout, sizeof(sout));

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += 2;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col16_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;
	unsigned char r0, r1, g0, g1, b0, b1;

	char *inp, *ninp, *outp;
	unsigned short sin, sinp, sout;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		/* FIXME: This will probably break as byte ordering changes */
		/* TODO: Don't use memcpy (all we need is
		 * assign and shift? */
		memcpy(&sin, inp, sizeof(sin));
		memcpy(&sinp, ninp, sizeof(sinp));

		r0 = get_red(sin);
		r1 = get_red(sinp);
		g0 = get_green(sin);
		g1 = get_green(sinp);
		b0 = get_blue(sin);
		b1 = get_blue(sinp);

		rintensity = (inleft * r0) + ((FP_ONE-inleft) * r1);
		gintensity = (inleft * g0) + ((FP_ONE-inleft) * g1);
		bintensity = (inleft * b0) + ((FP_ONE-inleft) * b1);

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			/* FIXME: This will probably break as byte ordering 
			 * changes */
			sout = 0;
			put_red(sout, CLIP_UP(
					fp_out(fp_mult(racc, fpscale)), 
					MAX_GRAY));
			put_green(sout, CLIP_UP(
					fp_out(fp_mult(gacc, fpscale)), 
					MAX_GRAY));
			put_blue(sout, CLIP_UP(
					fp_out(fp_mult(bacc, fpscale)), 
					MAX_GRAY));
			/* TODO: Don't use memcpy (all we need is
			 * assign and shift?) */
			memcpy(outp, &sout, sizeof(sout));

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_row32_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;

	register char *inp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	outp = out;

	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
			inp -= 4;
		}
		/* use linear interpolation for reconstruction */
		rintensity = inleft * inp[0] + 
			(FP_ONE-inleft) * inp[4];
		gintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * inp[5];
		bintensity = inleft*inp[2] + 
			(FP_ONE-inleft) * inp[6];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
			inp += 4;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			outp[0] = CLIP_UP(
				fp_out(fp_mult(racc, fpscale)), MAX_GRAY);
			outp[1] = CLIP_UP(
				fp_out(fp_mult(gacc, fpscale)), MAX_GRAY);
			outp[2] = CLIP_UP(
				fp_out(fp_mult(bacc, fpscale)), MAX_GRAY);

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp+= 4;
		}
	}

	return T_OK;
}

static int thinc_fant_simple_col32_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int rintensity, gintensity, bintensity;
	unsigned int racc, gacc, bacc;

	char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	racc = gacc = bacc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		rintensity = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];
		gintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * ninp[1];
		bintensity = inleft*inp[2] + 
			(FP_ONE-inleft) * ninp[2];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			racc += fp_mult(rintensity, inleft); 
			gacc += fp_mult(gintensity, inleft);
			bacc += fp_mult(bintensity, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			racc += fp_mult(rintensity, outleft); 
			gacc += fp_mult(gintensity, outleft);
			bacc += fp_mult(bintensity, outleft);

			outp[0] = CLIP_UP(
					fp_out(fp_mult(racc, fpscale)), 
					MAX_GRAY);
			outp[1] = CLIP_UP(
					fp_out(fp_mult(gacc, fpscale)), 
					MAX_GRAY);
			outp[2] = CLIP_UP(
					fp_out(fp_mult(bacc, fpscale)), 
					MAX_GRAY);

			racc = gacc = bacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}

	return T_OK;
}

/* UYVY */
int thinc_resize_uyvy(
	char *in, char *out,
	unsigned int inw, unsigned int inh,
	unsigned int outw, unsigned int outh,
	unsigned int instride, unsigned int outstride)
{
	return T_OK;
}

/* YUY2 */
static void thinc_fant_simple_row_yuy2_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlend, unsigned int fpscale,
	unsigned int fpinvscale);
static void thinc_fant_simple_col_yuy2_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, 
	unsigned int fpscale, unsigned int fpinvscale);

int thinc_resize_yuy2(
	char *in, char *out,
	unsigned int inw, unsigned int inh,
	unsigned int outw, unsigned int outh,
	unsigned int instride, unsigned int outstride)
{
	register unsigned int i;
	register char *inc, *outc;
	char *tmp;
	unsigned int tmplen, inlen, outlen, tmpstride;
	int free_tmp = T_NO;
	double scale, invscale;
	unsigned int fpscale, fpinvscale;
	
	inlen = (inw<<1)*inh;
	outlen = (outw<<1)*outh;

	/* rows first */
	if(inw == outw) {
		tmp = in;
		tmplen = inlen;
		tmpstride = instride;
	} else {
		tmpstride = outstride;
		tmplen = tmpstride*inh;
		tmp = malloc(tmplen);
		if(!tmp)
			return T_ERROR;
		free_tmp = T_YES;
		inc = in;
		outc = tmp;
		scale = (double)outw/(double)inw;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		for(i=0; i<inh; ++i) {
			thinc_fant_simple_row_yuy2_fp(inc, outc, inw, outw, 
				fpscale, fpinvscale);
			inc += instride;
			outc += tmpstride;
		}
	}
	/* columns */
	if(inh == outh) {
		/* TODO: Why not make out be tmp right away??? */
		memcpy(out, tmp, tmplen);
	} else {
		inc = tmp;
		outc = out;
		scale = (double)outh/(double)inh;
		invscale = 1.0/scale;

		fpscale = fp_in(scale);
		fpinvscale = fp_in(invscale);

		outw >>= 1;
	
		for(i=0; i<outw; ++i) {
			thinc_fant_simple_col_yuy2_fp(inc, outc, inh, outh, 
				tmpstride, outstride, fpscale, fpinvscale);
			inc += 4;	/* each macropixel is 4 bytes */
			outc += 4;
		}
	}

	if(free_tmp)
		free(tmp);

	return T_OK;
}

/* resize one row of yuy2, a packed format where each pixel looks like:
 * [ YO UO Y1 VO ]
 */
static void thinc_fant_simple_row_yuy2_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, unsigned int fpscale,
	unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int yintensity1, yintensity2, uintensity, vintensity;
	unsigned int yacc1, yacc2, uacc, vacc;

	register char *inp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	yacc1 = yacc2 = uacc = vacc = 0;

	inp = in;
	outp = out;

	inlen >>= 1;
	outlen >>= 1;
	
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index */
			--u;
			inp -= 4;
		}
		/* use linear interpolation for reconstruction */
		yintensity1 = inleft * inp[0] + 
			(FP_ONE-inleft) * inp[2];
		uintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * inp[5];
		yintensity2 = inleft * inp[2] + 
			(FP_ONE-inleft) * inp[4];
		vintensity = inleft*inp[3] + 
			(FP_ONE-inleft) * inp[7];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			yacc1 += fp_mult(yintensity1, inleft); 
			yacc2 += fp_mult(yintensity2, inleft); 
			uacc += fp_mult(uintensity, inleft);
			vacc += fp_mult(vintensity, inleft);

			outleft -= inleft; /*inleft portion has been filled*/
			inleft = FP_ONE;
			++u;
			inp += 4;
		} else {
			/* inleft >= outleft: input pixel not entirely 
			 * consumed before output pixel */
			yacc1 += fp_mult(yintensity1, outleft); 
			yacc2 += fp_mult(yintensity2, outleft); 
			uacc += fp_mult(uintensity, outleft);
			vacc += fp_mult(vintensity, outleft);

			outp[0] = CLIP_UP(
				fp_out(fp_mult(yacc1, fpscale)), MAX_GRAY);
			outp[1] = CLIP_UP(
				fp_out(fp_mult(uacc, fpscale)), MAX_GRAY);
			outp[2] = CLIP_UP(
				fp_out(fp_mult(yacc2, fpscale)), MAX_GRAY);
			outp[3] = CLIP_UP(
				fp_out(fp_mult(vacc, fpscale)), MAX_GRAY);

			yacc1 = yacc2 = uacc = vacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp+= 4;
		}
	}
}

static void thinc_fant_simple_col_yuy2_fp(char *in, char *out,
	unsigned int inlen, unsigned int outlen, 
	unsigned int instride, unsigned int outstride, 
	unsigned int fpscale, unsigned int fpinvscale)
{
	unsigned int u,x;
	unsigned int inleft, outleft;

	unsigned int yintensity1, yintensity2, uintensity, vintensity;
	unsigned int yacc1, yacc2, uacc, vacc;

	char *inp, *ninp, *outp;

	outleft = fpinvscale;
	inleft = FP_ONE;
	yacc1 = yacc2 = uacc = vacc = 0;

	inp = in;
	ninp = in+instride;
	outp = out;
	for(x=u=0; x<outlen; ) {
		if(u+1 == inlen) {
			/* XXX: HACK, HACK. Not too sure why we're overrunning
			 * this index. just go back one line and look the
			 * other way (when magnifying it does produce some
			 * repetition artifacts) */
			--u;
			ninp = inp;
			inp -= instride;
		}
		/* use linear interpolation for reconstruction */
		yintensity1 = inleft * inp[0] + 
			(FP_ONE-inleft) * ninp[0];
		uintensity = inleft * inp[1] + 
			(FP_ONE-inleft) * ninp[1];
		yintensity2 = inleft * inp[2] + 
			(FP_ONE-inleft) * ninp[2];
		vintensity = inleft*inp[3] + 
			(FP_ONE-inleft) * ninp[3];

		/* inleft < outleft: input pixel entirely consumed 
		 * before output pixel */
		if(inleft < outleft) {
			/*weigthed contribution*/
			yacc1 += fp_mult(yintensity1, inleft); 
			yacc2 += fp_mult(yintensity2, inleft); 
			uacc += fp_mult(uintensity, inleft);
			vacc += fp_mult(vintensity, inleft);
			outleft -= inleft; /*inleft portion has been filled*/

			inleft = FP_ONE;
			++u;
			inp = ninp;
			ninp += instride;
		} else {
			/* inleft >= outleft: input pixel not entirely consumed
			 * before output pixel */
			yacc1 += fp_mult(yintensity1, outleft); 
			yacc2 += fp_mult(yintensity2, outleft); 
			uacc += fp_mult(uintensity, outleft);
			vacc += fp_mult(vintensity, outleft);

			outp[0] = CLIP_UP(
					fp_out(fp_mult(yacc1, fpscale)), 
					MAX_GRAY);
			outp[1] = CLIP_UP(
					fp_out(fp_mult(uacc, fpscale)), 
					MAX_GRAY);
			outp[2] = CLIP_UP(
					fp_out(fp_mult(yacc2, fpscale)), 
					MAX_GRAY);
			outp[3] = CLIP_UP(
					fp_out(fp_mult(vacc, fpscale)), 
					MAX_GRAY);

			yacc1 = yacc2 = uacc = vacc = 0;
			inleft -= outleft;
			outleft = fpinvscale;
			++x;
			outp += outstride;
		}
	}
}
