/*
 * Application to test X draw operations.
 *
 * Released to the public domain. You may use this
 * code any way you wish, private, educational, or commercial.  It's free.
 *
 * Authors:
 * Leo Kim - Network Computing Lab - Columbia University 
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>		/* for va_args */
#include <time.h>		/* for "struct tm" in msg */
#include <unistd.h>		/* for sleep() */
#include <fcntl.h>		/* for open() */
#include <string.h>		/* for strerror() */
#include <sys/mman.h>		/* for mmap() */
#include <netinet/in.h> 	/* for ntohl() */
#include <errno.h>

#include <getopt.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

/* Defaults */
#define TEST_APP_NAME		"Test Draw"
#define TEST_DEPTH		24
#define TEST_WIDTH		600
#define TEST_HEIGHT		400
#define BITMAP_WIDTH		32
#define BITMAP_HEIGHT		32
#define TEST_SLEEP_DEFAULT	3

/* Colors */
#define TEST_FILL_COLOR		0xCC0000
#define TEST_GLYPH_COLOR	0x00CC00
#define TEST_BILEVEL_FG		0x999999
#define TEST_BILEVEL_BG		0xFFFFFF

/* Image files */
#define TEST_RAW_IMAGE1		"algore.raw"
#define TEST_RAW_IMAGE2		"angie.raw"
#define TEST_RAW_IMAGE3		"poland.raw"

/* Return values */
#define TEST_OKAY		0
#define TEST_ERROR		-1

#define DO_TIMESTAMPS
#define USE_TSC

#ifdef DO_TIMESTAMPS
# ifdef USE_TSC
#  include <asm/msr.h>
#  define gettime_tsc(t) { 	\
	rdtscl((t));		\
   }
#  define ts(x11) gettime_tsc(x11->tv[x11->curr++])
# else
#  include <sys/time.h>
#  include <time.h>
#  define ts(tv)	gettimeofday((tv), NULL)
# endif /* USE_TSC */
#else
# define ts(tv)
#endif	/* DO_TIMESTAMPS */

#define DO_SYNC

#ifdef DO_SYNC
#define SYNC()	XSync(x11->display, False)
#else
#define SYNC()
#endif /* DO_SYNC */

struct test_options {
	int image:1;
	int copy:1;
	int solid:1;
	int pixmap:1;
	int glyph:1;
	int bilevel:1;
	int use_shm:1;
	int verbose:1;
	int time:1; /* we always do it, this just controls whether we
			    print the numbers or not */
	unsigned int sleep;
	unsigned int width;
	unsigned int height;
};

#define option(x11, opt) (x11.opts.opt)
#define optionp(x11, opt) (x11->opts.opt)

#define verb(x11, m) if(option(x11, verbose)) msg(m);

/* Structures */
struct test_x11 {
	Display *display;
	Window window;
	GC gc;
	Pixmap pixmap;
	Visual *visual;
	XImage *image;
        XShmSegmentInfo shminfo;
	struct test_options opts;
	/* timestamps */
	unsigned long *tv;
	unsigned int num_tests;
	unsigned int curr;
};

/* Initialization prototypes */
static int initialize_x(struct test_x11 *x11);
static int createWindow(struct test_x11 *x11);
static int getVisual(struct test_x11 *x11);
static int createGC(struct test_x11 *x11);
static int createImage(struct test_x11 *x11, unsigned char *data, int width, 
    int height, int depth, int format);
static void destroyImage(struct test_x11 *x11);
static int createShmImage(struct test_x11 *x11, unsigned char *data, 
	int width, int height, int depth, int format);
static void shmDestroyImg(struct test_x11 *x11);

/* Test functions */
static int testRawImage(struct test_x11 *x11, char *file);
static int testCopy(struct test_x11 *x11);
static int testSolidFill(struct test_x11 *x11);
static int testGlyph(struct test_x11 *x11);
static int testBilevel(struct test_x11 *x11);
static int testPixmap(struct test_x11 *x11);

/* Utility functions */
static void print_times(struct test_x11 *x11);
static void cleanup(struct test_x11 *x11);
static void msg(const char *fmt, ...);

static int process_options(struct test_options *opts, int argc, char **argv);
static void usage(const char *prog_name);

/****************************** Main function ********************************/
int main(int argc, char **argv) 
{
	struct test_x11 x11;
	int retval = EXIT_SUCCESS;
	char c;

	if(process_options(&x11.opts, argc, argv) == TEST_ERROR)
		return EXIT_FAILURE;

	/* init values */
	if (! (x11.display = XOpenDisplay(NULL)) ){
		msg("Could not initialize display %s. Bailing...\n.", 
			XDisplayName(NULL));
		retval = EXIT_FAILURE;
		goto cleanup;
	}

	if(initialize_x(&x11) == TEST_ERROR) {
		retval = EXIT_FAILURE;
		goto cleanup;
	}

	x11.num_tests = 0;

	/* wait for the user to tell us it's ok to start */
	verb(x11, "ok, i'm testing:\n");
	if option(x11, image) {
		x11.num_tests+=3;
		verb(x11, "put images\n");
	}
	if option(x11, copy) {
		++x11.num_tests;
		verb(x11, "copies\n");
	}
	if option(x11, solid) {
		++x11.num_tests;
		verb(x11, "solid fills\n");
	}
	if option(x11, glyph) {
		++x11.num_tests;
		verb(x11, "glyphs\n");
	}
	if option(x11, bilevel) {
		++x11.num_tests;
		verb(x11, "bilevels\n");
	}
	if option(x11, pixmap) {
		++x11.num_tests;
		verb(x11, "pixmap fills\n");
	}

	x11.tv = malloc(x11.num_tests * 2 * sizeof(unsigned long));
	if(!x11.tv) {
		msg("OUT OF MEMORY!! RUN FOR YOUR LIFE!!\n");
		retval = EXIT_FAILURE;
		goto cleanup;
	}
again:
	x11.curr = 0;

	printf("press any key to start the tests...");
	c = getc(stdin);
	printf("here we go. in 2 seconds\n");

	/* don't start right away... */
	sleep(2);
	
	/* start the tests */
	if option(x11, image) {
		verb(x11, "put image test 1\n");
		if (TEST_ERROR == testRawImage(&x11, TEST_RAW_IMAGE1)) {
			msg("An error occured during RawImage on file %s. "
				"Bailing...\n", TEST_RAW_IMAGE1);
			retval = EXIT_FAILURE;
			goto cleanup;
		}
		sleep(option(x11, sleep));

		verb(x11, "put image test 2\n");
		if (TEST_ERROR == testRawImage(&x11, TEST_RAW_IMAGE2)) {
			msg("An error occured during RawImage on file %s. "
				"Bailing...\n", TEST_RAW_IMAGE2);
			retval = EXIT_FAILURE;
			goto cleanup;
		}
		sleep(option(x11, sleep));
		verb(x11, "put image test 3\n");
		if (TEST_ERROR == testRawImage(&x11, TEST_RAW_IMAGE3)) {
			msg("An error occured during RawImage on file %s. "
				"Bailing...\n", TEST_RAW_IMAGE3);
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, copy) {
		sleep(option(x11, sleep)); 
		verb(x11, "copy test\n");
		if (TEST_ERROR == testCopy(&x11)) {
			msg("An error occured during Copy. Bailing...\n");
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, solid) {
		sleep(option(x11, sleep)); 
		verb(x11, "solid fill test 1\n");
		if (TEST_ERROR == testSolidFill(&x11)) {
			msg("An error occured during SolidFill. Bailing...\n");
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, glyph) {
		sleep(option(x11, sleep));
		verb(x11, "glyph test 1\n");
		if (TEST_ERROR == testGlyph(&x11)) {
			msg("An error occured during testGlyph. Bailing...\n");
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, bilevel) {
		sleep(option(x11, sleep));
		verb(x11,"bilevel test 1\n");
		if (TEST_ERROR == testBilevel(&x11)) {
			msg("An error occured during testBilevel. "
				"Bailing...\n");
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, pixmap) {
		sleep(option(x11, sleep));
		verb(x11, "pixmap fill test 1\n");
		if (TEST_ERROR == testPixmap(&x11)) {
			msg("An error occured during testPixmap. Bailing...\n");
			retval = EXIT_FAILURE;
			goto cleanup;
		}
	}

	if option(x11, time)
		print_times(&x11);

	printf("\ni'm done. press 'r' to restart or any other key to "
		"finish...");
	c = getc(stdin);
	if(c == 'r') {
		c = getc(stdin);
		goto again;
	}
	printf("au revoir!\n");

cleanup:
	cleanup(&x11);
	return retval;
}

static void print_times(struct test_x11 *x11)
{
	unsigned int i;
	unsigned long start, end;

	printf("\n==== timing results ===\n");
#ifdef USE_TSC
	printf("test\tstart\t\tend\t\ttotal (cycles)\n");
#else
	printf("test\tstart\t\tend\t\ttotal (msec)\ttotal (sec)\n");
#endif
	for(i=0; i<x11->curr; ++i) {
#ifdef USE_TSC
		start = x11->tv[i];
		++i;
		end = x11->tv[i];
		printf("%d\t%lu\t%lu\t%lu\n", 
			(i/2)+1, start, end, end - start);
#else
		start = (x11->tv[i].tv_sec*1000) + (x11->tv[i].tv_usec/1000);
		++i;
		end = (x11->tv[i].tv_sec*1000) + (x11->tv[i].tv_usec/1000);
		printf("%d\t%lu\t%lu\t%10lu\t%lu\n", (i/2)+1, start, end, 
			end - start, (end-start)/1000);
#endif
	}
}

/**************************** Init functions *********************************/
static int process_options(struct test_options *opts, int argc, char **argv)
{
	struct option long_options[] = {
		/* name, has_arg, flag, val */
		{"help", 0, 0, 'H'},
		{"no-image", 0, 0, 'i'},
		{"no-copy", 0, 0, 'c'},
		{"no-solid", 0, 0, 's'},
		{"no-pixmap", 0, 0, 'p'},
		{"no-glyph", 0, 0, 'g'},
		{"no-bilevel", 0, 0, 'b'},
		{"use-shm", 0, 0, 'u'},
		{"sleep", 1, 0, 't'},
		{"width", 1, 0, 'w'},
		{"height", 1, 0, 'h'},
		{"verbose", 0, 0, 'v'},
		{"time", 0, 0, 'T'},
		{"all", 0, 0, 'a'},
		{0,0,0,0}
	};
	int opt, index=0;
	int err = TEST_OKAY;
	int any_test=0;

	/* set default options */
	opts->image = 0;
	opts->copy = 0;
	opts->solid = 0;
	opts->pixmap = 0;
	opts->glyph = 0;
	opts->bilevel = 0;
	opts->use_shm = 0;
	opts->verbose = 0;
	opts->time = 0;
	opts->sleep = TEST_SLEEP_DEFAULT;
	opts->width = TEST_WIDTH;
	opts->height = TEST_HEIGHT;

	while(1) {
		opt = getopt_long(argc, argv, "Hicspgbut:w:h:vTa", 
			long_options, &index);
		if(opt == -1)
			/* finished */
			break;

		switch(opt) {
			case 'H':
				/* not an error but who cares */
				err = TEST_ERROR;
				break;
			case 'a':
				any_test=1;
				opts->image = 1;
				opts->copy = 1;
				opts->solid = 1;
				opts->pixmap = 1;
				opts->glyph = 1;
				opts->bilevel = 1;
				break;
			case 'i':
				any_test=1;
				opts->image = 1;
				break;
			case 'c':
				any_test=1;
				opts->copy = 1;
				break;
			case 's':
				any_test=1;
				opts->solid = 1;
				break;
			case 'p':
				any_test=1;
				opts->pixmap = 1;
				break;
			case 'g':
				any_test=1;
				opts->glyph = 1;
				break;
			case 'b':
				any_test=1;
				opts->bilevel = 1;
				break;
			case 'u':
				opts->use_shm = 1;
				break;
			case 't':
				opts->sleep = atoi(optarg);
				break;
			case 'w':
				opts->width = atoi(optarg);
				break;
			case 'h':
				opts->height = atoi(optarg);
				break;
			case 'v':
				opts->verbose = 1;
				break;
			case 'T':
				opts->time = 1;
				break;
			default:
				err = TEST_ERROR;
				break;
		}
		if(err == TEST_ERROR)
			break;
	}

	if(!any_test) {
		fprintf(stderr, "you didn't select any tests to run!\n\n");
		err = TEST_ERROR;
	}

	if(err == TEST_ERROR) 
		usage(argv[0]);

	return err;
}

static void usage(const char *prog_name) 
{
	fprintf(stderr,
		"usage: %s [options]\n"
		"options:\n"
		"   --image/-i		test putimage\n"
		"   --copy/-c		test copy\n"
		"   --solid/-s		test solid fill\n"
		"   --pixmap/-p		test pixmap fill\n"
		"   --glyph/-g		test glyphs\n"
		"   --bilevel/-b	test bilevels\n"
		"   --all/-a		test everything\n"
		"   --use-shm/-u	use shared memory\n"
		"   --sleep/-t		how long to sleep between tests\n"
		"   --width/-w	        width of window\n"
		"   --height/-H         height of window\n"
		"   --verbose/-v        be verbose\n"
		"   --time/-T      print time taken by each test\n"
		, prog_name
	);
}

static int initialize_x(struct test_x11 *x11)
{
	x11->window = (Window)-1;
	x11->gc = (GC)-1;
	x11->pixmap = (Pixmap)-1;
	x11->visual = NULL;
	x11->image = NULL;

	/* initialize visual */
	if (TEST_ERROR == getVisual(x11)) {
		msg("Could not init Visual.\n");
		return TEST_ERROR;
	}

	/* create window */
	if (TEST_ERROR == createWindow(x11)) {
		msg("Could not create Window.\n");
		return TEST_ERROR;
	}

	/* create GC */
	if (TEST_ERROR == createGC(x11)) {
		msg("Could not create graphics context.\n");
		return TEST_ERROR;
	}

	/* ok, map the window and we're all set */
	XMapWindow(x11->display, x11->window);
	XFlush(x11->display);

	return TEST_OKAY;
}

static int createWindow(struct test_x11 *x11)
{
	const char *func = __func__;
	XSizeHints *xsh;
	XClassHint *xch;
	XWMHints *xwmh; 
	XTextProperty wname, iname;
	XSetWindowAttributes xswa;
	unsigned long fgpix, bgpix;
	char *appname = TEST_APP_NAME;

	bgpix = BlackPixel(x11->display, DefaultScreen(x11->display));
	fgpix = WhitePixel(x11->display, DefaultScreen(x11->display));

	if (NULL == (xsh = XAllocSizeHints())) {
		msg("%s: Error allocating size hints\n", func);
		return TEST_ERROR;
	} 
	xsh->flags = PSize | PMaxSize | PMinSize;
	xsh->min_width = xsh->max_width = xsh->width = optionp(x11, width);
	xsh->min_height = xsh->max_height = xsh->height = optionp(x11, height);

	/* create the window */
	x11->window = XCreateSimpleWindow(x11->display, 
		DefaultRootWindow(x11->display),
		0, 0, /* initial position */
		xsh->width, xsh->height,
		0, /* border width */
		fgpix, bgpix);

	if (NULL == (xch = XAllocClassHint())) {
		msg("%s: Error allocating class hint\n", func);
		return TEST_ERROR;
	}

	xch->res_name = appname;
	xch->res_class = appname;

	/* window and icon name */
	if (XStringListToTextProperty(&appname, 1, &wname) == 0 ||
		XStringListToTextProperty(&appname, 1, &iname) == 0) 
	{
		msg("%s: Error creating XTextProperty\n", func);
		return TEST_ERROR;
	}

	/* XXX: i think this is deprecated in the x world... one day we'll
	 * fix it */
	if (NULL == (xwmh = XAllocWMHints())) {
		msg("%s: Error allocating WM hints\n", func);
		return TEST_ERROR;
	}

	xwmh->flags = (InputHint|StateHint);
	xwmh->input = True;
	xwmh->initial_state = NormalState;
	XSetWMProperties(x11->display, x11->window, &wname, &iname, NULL, 0,
		xsh, xwmh, xch);

	xswa.colormap = DefaultColormap(x11->display, 
		DefaultScreen(x11->display)); 
	xswa.bit_gravity = CenterGravity;

	XChangeWindowAttributes(x11->display, x11->window,
		(CWColormap|CWBitGravity), &xswa);

	return TEST_OKAY;
}

static int getVisual(struct test_x11 *x11)
{
	const char *func = __func__;
	XVisualInfo tmpl, *vinfo;
	int nvis, mask;

	tmpl.screen = DefaultScreen(x11->display);
	tmpl.class = TrueColor;
	tmpl.depth = TEST_DEPTH;
	mask = VisualScreenMask | VisualClassMask | VisualDepthMask;
	vinfo = XGetVisualInfo(x11->display, mask, &tmpl, &nvis);
	if(!vinfo) {
		msg("%s: Couldn't find suitable x11->visual\n", func);
		return TEST_ERROR;
	}
	x11->visual = vinfo[0].visual;
	XFree(vinfo);

	return TEST_OKAY;
}

static int createGC(struct test_x11 *x11)
{
	unsigned long bgpix, fgpix;
	XGCValues gcv;

	bgpix = BlackPixel(x11->display, DefaultScreen(x11->display));
	fgpix = WhitePixel(x11->display, DefaultScreen(x11->display));

	gcv.foreground = fgpix;
	gcv.background = bgpix;

	x11->gc = XCreateGC(x11->display, x11->window,
	(GCForeground | GCBackground), &gcv);

	return TEST_OKAY;
}

static int createImage(struct test_x11 *x11, unsigned char *data, int width, 
    int height, int depth, int format)
{
	int offset = 0;
	int bitmap_pad = 32;
	int bytes_per_line = 0; /* # of bytes between start of successive scan
				   lines. 0 -> lines are contiguous */

	x11->image = XCreateImage(x11->display, x11->visual, depth, format, 
		offset, data, width, height, bitmap_pad, bytes_per_line);

	if(!x11->image) return TEST_ERROR;

	x11->image->byte_order = LSBFirst;

	return TEST_OKAY;
}

static void destroyImage(struct test_x11 *x11)
{
	x11->image->data = NULL;
	XDestroyImage(x11->image);
	x11->image = NULL;
}

static int createShmImage(struct test_x11 *x11, unsigned char *data, 
	int width, int height, int depth, int format)
{
	/* do we have SHM? */
	if(!XShmQueryExtension(x11->display)) {
		return TEST_ERROR;
	}

	/* create shm x11->image */
	x11->image = XShmCreateImage(x11->display, x11->visual, depth, 
		format, NULL, &x11->shminfo, width, height);
	if(!x11->image) {
		msg("couldn't create shm image\n");
		return TEST_ERROR;
	}

	/* allocate memory for it */
	x11->shminfo.shmid = shmget (IPC_PRIVATE, 
		x11->image->bytes_per_line * x11->image->height, 
		IPC_CREAT|0777);
	if(x11->shminfo.shmid == -1) {
		msg("couldn't get shmid\n");
		XDestroyImage(x11->image);
		return TEST_ERROR;
	}
	/* attach memory to us */
	x11->shminfo.shmaddr = x11->image->data = shmat(x11->shminfo.shmid, 
		0, 0);
	if(x11->shminfo.shmaddr == (char*)-1) {
		msg("couldn't attach to shm\n");
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->image);
		return TEST_ERROR;
	}
	/* if we set it to True then GetImage's won't work.. do we
	 * care? */
	x11->shminfo.readOnly = True;

	/* tell server to attach the memory segment */
	if(XShmAttach(x11->display, &x11->shminfo)  == 0) {
		msg("couldn't x11 shm attach\n");
		shmdt(x11->shminfo.shmaddr);
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->image);
		return TEST_ERROR;
	}
	XSync(x11->display, False);

	/* now copy the data to the image */
	memcpy(x11->image->data, data,
		x11->image->bytes_per_line * x11->image->height);

	return TEST_OKAY;
}

static void shmDestroyImg(struct test_x11 *x11)
{
	XShmDetach(x11->display, &x11->shminfo);
	shmdt(x11->shminfo.shmaddr);
	shmctl(x11->shminfo.shmid, IPC_RMID, 0);
	x11->image->data = NULL;
	XFree(x11->image);
	x11->image = NULL;
}

/************************** Test functions *********************************/
static int testRawImage(struct test_x11 *x11, char *file)
{
	int fd;
	char *data;
	int width, height, bpp, size, depth;
	int retval = TEST_OKAY;

	if(-1 == (fd = open(file, O_RDONLY))) {
		msg("Could not open input file %s. %s\n", strerror(errno));
		retval = TEST_ERROR;
		goto end_raw;
	} 

	read(fd, &width, 4);
	read(fd, &height, 4);
	read(fd, &bpp, 4);
	width = ntohl(width);
	height = ntohl(height);
	bpp = ntohl(bpp);
	size = width*height*bpp;

	bpp *= 8;
	depth = (bpp == 32 ? 24 : bpp);

	data = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0);
	if((void*)-1 == data) {
		msg("Could not map input file: %s\n", strerror(errno));
		close(fd);
		retval = TEST_ERROR;
		goto end_raw;
	}

	if (optionp(x11, use_shm)) {
		retval = createShmImage(x11, data, width, height, depth, 
			ZPixmap);
	} else {
		retval = createImage(x11, data, width, height, depth, 
			ZPixmap);
	}
	
	SYNC();
	ts(x11);
	if(TEST_ERROR == retval)
		goto cleanup_raw;

	XPutImage(x11->display, x11->window, x11->gc, x11->image, 
		10, 10, 0, 0, width-20, height-20);
	SYNC();
	ts(x11);

cleanup_raw:
	munmap(data, size);
	if (optionp(x11, use_shm)) 
		shmDestroyImg(x11);
	else
		destroyImage(x11);
	close(fd);

end_raw:
	return retval;
}

static int testCopy(struct test_x11 *x11)
{
	SYNC();
	ts(x11);
	XCopyArea(x11->display, x11->window, x11->window, x11->gc, 0,
		0, optionp(x11, width) >> 1, optionp(x11,height), 
		optionp(x11,width) >> 1, 0);
	SYNC();
	ts(x11);

	return TEST_OKAY;
}

static int testSolidFill(struct test_x11 *x11)
{
	XGCValues gcv;

	gcv.foreground = TEST_FILL_COLOR;
	gcv.fill_style = FillSolid;
	XChangeGC(x11->display, x11->gc, (GCForeground | GCFillStyle), &gcv);

	SYNC();
	ts(x11);
	XFillRectangle(x11->display, x11->window, x11->gc, 0, 0, 
		optionp(x11,width), optionp(x11,height));
	SYNC();
	ts(x11);

	return TEST_OKAY;
}

static int testBilevel(struct test_x11 *x11)
{
	XGCValues gcv;
	char *bitmap;
	int size = (BITMAP_WIDTH * BITMAP_HEIGHT) >> 3;

	/* The stipple */
	if ((Pixmap)-1 != x11->pixmap) {
		XFreePixmap(x11->display, x11->pixmap);
	}

	if (NULL == (bitmap = (calloc(size, sizeof(unsigned char))))) {
		msg("Error: could not allocate enough memory for bitmap.\n");
		return TEST_ERROR;
	}

	memset(bitmap, 1, size);

	x11->pixmap = XCreateBitmapFromData(x11->display, x11->window,
		bitmap, BITMAP_WIDTH, BITMAP_HEIGHT);
	if(x11->pixmap == None) {
		msg("Error: Failed to create stipple\n");
		free(bitmap);
		return TEST_ERROR;
	}

	gcv.fill_style = FillOpaqueStippled;
	gcv.foreground = TEST_BILEVEL_FG;
	gcv.background = TEST_BILEVEL_BG;
	gcv.stipple = x11->pixmap;
	gcv.ts_x_origin = 0;
	gcv.ts_y_origin = 0;

	XChangeGC(x11->display, x11->gc, (GCFillStyle | GCStipple |
		GCForeground | GCTileStipXOrigin | GCTileStipYOrigin),
		&gcv);

	SYNC();
	ts(x11);
	XFillRectangle(x11->display, x11->window, x11->gc, 0, 0, 
		optionp(x11,width), optionp(x11, height)); 
	SYNC();
	ts(x11);

	free(bitmap);

	return TEST_OKAY;
}

static int testGlyph(struct test_x11 *x11)
{
	XGCValues gcv;
	char *bitmap;
	int size = (BITMAP_WIDTH * BITMAP_HEIGHT) >> 3;

	/* The stipple */
	if ((Pixmap)-1 != x11->pixmap) {
		XFreePixmap(x11->display, x11->pixmap);
	}

	if (NULL == (bitmap = (calloc(size, sizeof(unsigned char))))) {
		msg("Error: could not allocate enough memory for bitmap.\n");
		return TEST_ERROR;
	}
	memset(bitmap, 1, size);

	x11->pixmap = XCreateBitmapFromData(x11->display, x11->window,
		bitmap, BITMAP_WIDTH, BITMAP_HEIGHT);
	if(x11->pixmap == None) {
		msg("Error: Failed to create stipple\n");
		free(bitmap);
		return TEST_ERROR;
	}

	gcv.fill_style = FillStippled;
	gcv.foreground = TEST_GLYPH_COLOR;
	gcv.stipple = x11->pixmap;
	gcv.ts_x_origin = 0;
	gcv.ts_y_origin = 0;

	XChangeGC(x11->display, x11->gc, (GCFillStyle | GCStipple |
		GCForeground | GCTileStipXOrigin | GCTileStipYOrigin),
		&gcv);

	SYNC();
	ts(x11);
	XFillRectangle(x11->display, x11->window, x11->gc, 0, 0, 
		optionp(x11, width), optionp(x11,height)); 
	SYNC();
	ts(x11);

	free(bitmap);

	return TEST_OKAY;
}

static int testPixmap(struct test_x11 *x11) 
{
	XGCValues gcv;
	int fd;
	char *data;
	int width, height, bpp, size, depth;
	int retval = TEST_OKAY;

	/* open up al gore */
	if(-1 == (fd = open(TEST_RAW_IMAGE1, O_RDONLY))) {
		msg("Could not open input file %s. %s\n", strerror(errno));
		retval = TEST_ERROR;
		goto end_raw;
	} 

	read(fd, &width, 4);
	read(fd, &height, 4);
	read(fd, &bpp, 4);
	width = ntohl(width);
	height = ntohl(height);
	bpp = ntohl(bpp);
	size = width*height*bpp;

	bpp *= 8;
	depth = (bpp == 32 ? 24 : bpp);

	data = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0);
	if((void*)-1 == data) {
		msg("Could not map input file: %s\n", strerror(errno));
		close(fd);
		retval = TEST_ERROR;
		goto end_raw;
	}

	if (optionp(x11, use_shm)) {
		retval = createShmImage(x11, data, width, height, depth, 
			ZPixmap);
	} else {
		retval = createImage(x11, data, width, height, depth, 
			ZPixmap);
	}

	if(TEST_ERROR == retval)
		goto cleanup_raw;
	
	if ((Pixmap)-1 != x11->pixmap) {
		XFreePixmap(x11->display, x11->pixmap);
	}

	x11->pixmap = XCreatePixmap(x11->display, x11->window, width*2, 
		height*2, depth);

	XPutImage(x11->display, x11->pixmap, x11->gc, x11->image, 0, 0, 0, 0, 
		width, height);

	gcv.fill_style = FillTiled;
	gcv.tile = x11->pixmap;
	gcv.ts_x_origin = 0;
	gcv.ts_y_origin = 0;
	XChangeGC(x11->display, x11->gc, (GCFillStyle | GCTile
		| GCTileStipXOrigin | GCTileStipYOrigin ),
		&gcv);

	SYNC();
	ts(x11);
	XFillRectangle(x11->display, x11->window, x11->gc, 20, 4,
		width-20, height*2);
	SYNC();
	ts(x11);

cleanup_raw:
	munmap(data, size);
	close(fd);
	if (optionp(x11, use_shm)) 
		shmDestroyImg(x11);
	else
		destroyImage(x11);

end_raw:
	return retval;
}

/************************ Utility functions ********************************/
static void cleanup(struct test_x11 *x11) 
{
	if((GC)-1 != x11->gc) XFreeGC(x11->display, x11->gc);
	if(x11->image) {
		msg("BUG: somebody didn't cleanup the image!\n");
	}
	if((Window)-1 != x11->window) XDestroyWindow(x11->display, x11->window);
	if((Pixmap)-1 != x11->pixmap) XFreePixmap(x11->display, x11->pixmap);

	XCloseDisplay(x11->display);
}

static void msg(const char *fmt, ...)
{
	va_list v;
	time_t t;
	struct tm b;

	time(&t);	    
	localtime_r(&t, &b);

	va_start(v, fmt);
	vfprintf(stdout, fmt, v);
	va_end(v);
}
