/*
 * Utility to display framebuffers. Opens local file with fb contents and
 * displays them on an X11 window. Format of file:
 * width height depth bpp fb-contents
 * Note: this is a hack
 *
 * Released to the public domain. You may use this
 * code any way you wish, private, educational, or commercial.  It's free.
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>

#define min(x, y) (((x) < (y)) ? (x) : (y))

/* information about the fb */
struct fb_header {
	int width, height;
	int depth, bpp;
};

struct fb_info {
	struct fb_header h;
	unsigned char *data;
};

struct fb_info* read_data(const char *file);
int readData(int fd, void *buf, int count);

int main(int argc, char **argv)
{
	Window	Main;
	Display *disp;
	Pixmap thePixmap;

	/* image specific stuff */
	XImage	*theImage;
	Visual *visual;
	unsigned char *data;
	int format = ZPixmap;
	struct fb_info *info;
	unsigned int depth;
	unsigned int width, height;
	int offset = 0;
	int bitmap_pad = 8;
	int bytes_per_line = 0; /* # of bytes between start of successive scan
				   lines. 0 -> lines are contiguous */

	/* other necessary stuff */
	Colormap theColormap;
	XVisualInfo tmpl, *vinfo;
	Visual *theVisual;
	int nvis, mask;
	XSizeHints *xsh;
	XClassHint *xch;
	XWMHints *xwmh;
	XGCValues gcv;
	GC theGC;
	XEvent theEvent;
	XSetWindowAttributes xswa;
	XTextProperty wname, iname;
	unsigned long bgpix, fgpix;
	char *title;
	int Done = 0;
	char *display_name = NULL;

	if(argc != 2) {
		fprintf(stderr, "usage: %s file\n", 
			argv[0]);
		exit(1);
	}

	title = argv[1];

	info = read_data(argv[1]);
	if(!info) {
		fprintf(stderr, "%s: couldn't read image data\n", argv[0]);
		exit(1);
	}
	data = info->data;
	depth =info->h.depth;
	width = info->h.width;
	height = info->h.height;

	fprintf(stderr, "image is %dx%dx%d\n", width, height, depth);

	/********************* open display **********************/
	if((disp = XOpenDisplay(display_name)) == NULL) {
		fprintf(stderr, "%s: can't open display %s\n", argv[0],
			XDisplayName(display_name));
		exit(1);
	}

	/******************* set up colors ***************/
	/* here we look for a visual we support. Currently that's only
	 * TrueColor at 24bits */
	tmpl.screen = DefaultScreen(disp);
	tmpl.class = TrueColor;
	tmpl.depth = depth;
	mask = VisualScreenMask | VisualClassMask | VisualDepthMask;
	vinfo = XGetVisualInfo(disp, mask, &tmpl, &nvis);
	if(!vinfo) {
		fprintf(stderr, "%s: Couldn't find visual!!\n", argv[0]);
		exit(1);
	}
	
	theVisual = vinfo[0].visual;
	XFree(vinfo);

	theColormap = XCreateColormap(disp, DefaultRootWindow(disp),
			theVisual, AllocNone);
	bgpix = BlackPixel(disp, DefaultScreen(disp));
	fgpix = WhitePixel(disp, DefaultScreen(disp));

	/******************* initial size *******************/
	if((xsh = XAllocSizeHints()) == NULL) {
		fprintf(stderr, "%s: Error allocating size hints\n", argv[0]);
		exit(1);
	}
	/* FIXME: Need to implement scrollbar so we don't have min size */
	xsh->flags = PSize | PMaxSize | PMinSize;
	xsh->min_width = xsh->max_width = xsh->width = 
		min(width, DisplayWidth(disp, DefaultScreen(disp)));
	xsh->min_height = xsh->max_height = xsh->height = 
		min(height, DisplayHeight(disp, DefaultScreen(disp)));

	/***************** create the window ****************/
	Main = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
		0, 0, /* initial position */
		xsh->width, xsh->height, 
		0, /* border width */
		fgpix, bgpix);

	/************* wm properties **********************/
	if((xch = XAllocClassHint()) == NULL) {
		fprintf(stderr, "%s: Error allocating class hint\n", argv[0]);
		exit(1);
	}
	xch->res_name = argv[0];
	xch->res_class = argv[0];

	/* window and icon name */
	if( XStringListToTextProperty(&title, 1, &wname) == 0 ||
		XStringListToTextProperty(&title, 1, &iname) == 0) {
		fprintf(stderr, "%s: Error creating XTextProperty\n", argv[0]);
		exit(1);
	}

	if((xwmh = XAllocWMHints()) == NULL) {
		fprintf(stderr, "%s: Error allocating WM hints\n", argv[0]);
		exit(1);
	}

	xwmh->flags = (InputHint|StateHint);
	xwmh->input = True;
	xwmh->initial_state = NormalState;
	XSetWMProperties(disp, Main,&wname,&iname,argv,argc,xsh,xwmh,xch);

	/* the gc */
	gcv.foreground = fgpix;
	gcv.background = bgpix;

	theGC = XCreateGC(disp, Main, (GCForeground | GCBackground), &gcv);

	/******************* Input events for main window ************/
	XSelectInput(disp, Main, ExposureMask);

	/****************** Create the image ***************/
	visual = DefaultVisual(disp, DefaultScreen(disp));
	theImage = XCreateImage(disp, visual, depth, format, offset, data, 
		width, height, bitmap_pad, bytes_per_line);
	if(!theImage) {
		fprintf(stderr, "%s: Couldn't create fb image!!\n", argv[0]);
		exit(1);
	}

	printf("here\n");

	/* FIXME: should detect it */
	theImage->byte_order = LSBFirst;

	/* create pixmap and copy image to it */
	fprintf(stderr, "Image is %dx%d@%d\n", theImage->width, 
		theImage->height, theImage->depth);
	thePixmap = XCreatePixmap(disp, Main, theImage->width, theImage->height,
			theImage->depth);

	/*** set window attributes ****/

	/* not sure if we should install our own colormap... currently it
	 * works either way */
#if 0
	xswa.colormap = theColormap;
#else
	xswa.colormap = DefaultColormap(disp, DefaultScreen(disp));
#endif
	xswa.bit_gravity = CenterGravity;
	xswa.background_pixmap = thePixmap;
	XChangeWindowAttributes(disp, Main, 
		(CWColormap | CWBitGravity | CWBackPixmap), &xswa);


	XPutImage(disp, thePixmap, theGC, theImage, 0,0,0,0, theImage->width,
			theImage->height);
	/* image now on server, destroy our copy */
	XDestroyImage(theImage);
	
	/********** Map the window ***************/
	XMapWindow(disp, Main);

	/************ Event loop ****************/
	while(!Done) {
		XNextEvent(disp, &theEvent);
		if(theEvent.xany.window != Main)
			continue;
		/* FIXME: Seems like this is too slow because we're losing
		 * some expose events (for example when moving windows
		 * on top of ours we get a black trail)
		 * -> the fprintf() doesn't seem to be the reason */
		if(theEvent.type == Expose && theEvent.xexpose.count == 0) {
			/* redraw */
			XCopyArea(disp, thePixmap, Main, theGC, 
				theEvent.xexpose.x, theEvent.xexpose.y,
				theEvent.xexpose.width, theEvent.xexpose.height,
				theEvent.xexpose.x, theEvent.xexpose.y);
		}
	}

	/* the end */
	XFreeGC(disp, theGC);
	XFreePixmap(disp, thePixmap);
	XDestroyWindow(disp, Main);
	XCloseDisplay(disp);
	return 0;
}

/* gets data from host:port and returns size of data read. host can
 * be name or ip address. port can be numeric or name from /etc/services.
 * data is allocated according to how much data is read */
struct fb_info* read_data(const char *file)
{
	int count, got, buf_size;
	unsigned char *curr;
	struct fb_info* info;
	int f;

	f = open(file, O_RDONLY);
	if(f == -1) {
		fprintf(stderr, "Couldn't open file %s: %s\n", file,
			strerror(errno));
		return NULL;
	}

	info = calloc(1, sizeof(struct fb_info));
	if(!info) {
		perror("Couldn't read data");
		close(f);
		return NULL;
	}

	/* get header */
	if(read(f, &info->h, sizeof(struct fb_header)) == -1) {
		perror("Error reading data");
		free(info);
		close(f);
		return NULL;
	}

	printf("fb: %d,%d,%d,%d\n", info->h.width, info->h.height, 
		info->h.depth, info->h.bpp);

	buf_size = info->h.width*(info->h.bpp/8);

	/* alloc memory */
	info->data = calloc(buf_size*info->h.height, sizeof(unsigned char));
	if(!info->data) {
		perror("Error reading data");
		free(info);
		close(f);
		return NULL;
	}

	printf("reading %d bytes\n", info->h.height*buf_size);

	/* ok good, read data */
	curr = info->data;
	for(count = 0; count < info->h.height; count++) {
		got = readData(f, curr, buf_size);
		if(got != buf_size) { /* XXX: Does writeAll do the same? */
			perror("Error reading data");
			free(info->data);
			free(info);
			close(f);
			return NULL;
		}
		curr+=buf_size;
	}

	return info;
}

int readData(int fd, void *buf, int count)
{
	char *func = "ReadData()";
	int total, res;
	char *curr = (char *) buf;

	total = 0;
	do {
		res = read(fd, (void *) curr, count - total);
		if (res == -1) {
			if (errno == EWOULDBLOCK || errno == EAGAIN) {
				fprintf(stderr, "%s: no data to read\n", func);
				return -2;
			} else if (errno == EINTR) {
				fprintf(stderr, "%s: interrupted, trying "
					"again\n", func);
				continue;
			} else {
				fprintf(stderr, "Error reading: %s\n", 
					strerror(errno));
				return -1;
			}
		} 
		total += res;
		curr += res;
	} while (total != count);

	return total;
}
