#!/usr/bin/perl
####################################################################
# Summary: Hack Hack Hack
# Creates a Graph based on a THINC log file. Uses code from 
# thinc_log_parser.pl (by Leo Kim)
#
# Authors: 
# Ricardo A. Baratto - Network Computing Lab - Columbia University
#
# Released to the public domain. You may use this
# code any way you wish, private, educational, or commercial.  It's free.
####################################################################

use strict;
use GraphViz;
use Getopt::Std;

# global variables
use vars qw(
	$DEFAULT_LOG_FILE
	$DEFAULT_OUT_FILE
	@DRAW_FUNCS
	@CREATE_FUNCS
	@DESTROY_FUNCS
	@COPY_FUNCS
	@FUNCTION_END_SIGNIFIERS
	$G
	$PIX_COUNT
	$WIN_COUNT
	$COPY_COUNT
);

# state variables
# STATE holds the current state. STATE_ADD contains any additional information
# that needs to be passed to the STATE handling function.
use vars qw(
	$STATE
	$STATE_ADD
	$ST_INIT
	$ST_END
	$ST_NEWPIX
	$ST_NEWWIN
	$ST_COPY
);

$DEFAULT_LOG_FILE = "/var/log/XFree86.0.log";
$DEFAULT_OUT_FILE = "logGraph.png";

# functions we're interested in
@DRAW_FUNCS = (
	"thincPutImage",
	"thincCopyPlane",
	"thincPolyPoint",
	"thincPolylines",
	"thincPolySegment",
	"thincPolyRectangle",
	"thincPolyArc",
	"thincFillPolygon",
	"thincPolyFillRect",
	"thincPolyFillArc",
	"thincPolyText8",
	"thincPolyText16",
	"thincImageText8",
	"thincImageText16",
	"thincImageGlyphBlt",
	"thincPolyGlyphBlt",
	"thincFillSpans",
	"thincSetSpans",
	"thincPushPixels");
@CREATE_FUNCS = ("thincCreatePixmap", "thincCreateWindow");
@DESTROY_FUNCS = ("thincCreatePixmap", "thincCreateWindow"); 
@COPY_FUNCS = ("thincCopyArea");
@FUNCTION_END_SIGNIFIERS = ("finished", "not sending update", "error", "bug");

# option parsing
use vars qw(
	$opt_f
	$opt_o
	$opt_h
);

getopts('hf:o:');

main();

sub usage {
	print STDERR "One day it'll be written. For now, use the source luke\n";
	exit(0);
}

#### functions for each of the possible states ###

# Default state, looking for something interesting.
sub atInit {
	my $funcname;
	my @words;

	# get the function name
	@words = split(/[\s+]/);
	$funcname = $words[0];

	# clean up function name
	$funcname =~ s/[:.]//;

#	print "at INIT, $_\n";

	# check if we found something interesting
	if(isANewObject()) {
		my $newObj;

		# check what kind of new object. We need to call the functions right
		# away in case the info they need is in this line.
		if($funcname =~ /Pixmap/) {
			$STATE = $ST_NEWPIX;
			$STATE_ADD = $funcname;
			atNewPix();
		} elsif($funcname =~ /Window/) {
			$STATE = $ST_NEWWIN;
			$STATE_ADD = $funcname;
			atNewWin();
		} else {
			print STDERR "new object but don't know what kind\n";
		}
	} elsif(isACopy()) {
			$STATE = $ST_COPY;
			$STATE_ADD = $funcname;
			atCopy();
	} else {
#			print "operation not supported yet, sorry\n";
	}
}

# returns true if the function is one that creates a new object
sub isANewObject {
	my $pattern;
	foreach $pattern (@CREATE_FUNCS) {
		if(/$pattern/i) {
			return 1;
		}
	}
	return 0;
}

# returns true if the function is one that makes a copy
sub isACopy {
	my $pattern;
	foreach $pattern (@COPY_FUNCS) {
		if(/$pattern/i) {
			return 1;
		}
	}

	return 0;
}

### Looking for end of function
sub atEnd {
	my $func = "atEnd()";

	if(isAtFunctionEnd()) {
		if(/$STATE_ADD/) {
			print STDERR "$func: Found correct function end: $STATE_ADD\n";
			$STATE = $ST_INIT;
		}
	}
}

# returns true if a keyword is found in the @FUNCTION_END_SIGNIFIERS array
# false if not.
sub isAtFunctionEnd {
    my $pattern;

    foreach $pattern (@FUNCTION_END_SIGNIFIERS) {
        if (/$pattern/i) {
            return 1;
        }
    }
    return 0;
}


### New Pixmap
sub atNewPix {
	my $func = "atNewPix()";

	if($STATE != $ST_NEWPIX) {
		print STDERR "$func: Invalid state $STATE to be here\n";
		exit(0);
	}

	# look for ID in the line. If found, create a new node and go on to
	# ST_END (looking for the end of this function call
	if(/id (0x[0-9a-f]*)/i) {
		my $funcname;
		my @words;

#		print STDERR "$func: Found ID: $1. Creating new node\n";
#		$G->add_node($1, shape => 'ellipse');
		# STATE_ADD stays the same
		$STATE = $ST_END;
		$PIX_COUNT++;
		atEnd();
	}
}

### New Window
sub atNewWin {
	my $func = "atNewWin()";

	if($STATE != $ST_NEWWIN) {
		print STDERR "$func: Invalid state $STATE to be here\n";
		exit(0);
	}

	# look for ID in the line. If found, create a new node and go on to
	# ST_END (looking for the end of this function call
	if(/id (0x[0-9a-f]*)/i) {
#		print STDERR "$func: Found ID: $1. Creating new node\n";
#		$G->add_node($1, shape => 'box');
		# STATE_ADD has correct info
		$STATE = $ST_END;
		$WIN_COUNT++;
		atEnd();
	}
}

### Copy
# Using closures so we can have $src and $dst be persistent
{
	my $src = "";
	my $dst = "";
	
	sub atCopy {
		my $func = "atCopy()";

		if($STATE != $ST_COPY) {
				print STDERR "$func: Invalid state $STATE to be here\n";
				exit(0);
		}

		# need to look for src and dst
		if($src eq "") {
			if(/SRC/) {
				if(/id: (0x[0-9a-f]*)/i) {
					print STDERR "$func: found source. ID: $1\n";
					$src = $1;
					if(/Window/) {
						$G->add_node($src, shape => 'box');
					} else {
						$G->add_node($src, shape => 'ellipse');
					}
				}
			}
		} 

		if($dst eq "") {
			if(/DST/) {
				if(/id: (0x[0-9a-f]*)/i) {
					print STDERR "$func: found destination. ID: $1\n";
					$dst = $1;
					if(/Window/) {
						$G->add_node($dst, shape => 'box');
					} else {
						$G->add_node($dst, shape => 'ellipse');
					}
				}
			}
		}

		# are we finished?
		if(($src ne "") and ($dst ne "")) {
			# have both, create the edge
			print STDERR "$func: Creating edge $src => $dst\n";
#			$G->add_node($src);
#			$G->add_node($dst);
			$G->add_edge($src => $dst);
			# done, look for end of function call. No need to change
			# $ST_ADD
			$STATE = $ST_END;
			$COPY_COUNT++;
			$src = "";
			$dst = "";
			atEnd();
		}
	}
}

sub main {
	# only help?
	if($opt_h) {
		usage();
	}

	# local vars
	my @words;
	my @stack;
	my $funcname = "";
	my $ans;

	# initialize states
	$ST_INIT = 0;
	$ST_END = 1;
	$ST_NEWPIX = 2;
	$ST_NEWWIN = 3;
	$ST_COPY = 4;
	$STATE = $ST_INIT;

	# init counters
	$PIX_COUNT = 0;
	$WIN_COUNT = 0;
	$COPY_COUNT = 0;

	# the graph
	# boxes => windows
	# ellipse => pixmaps
#	$G = GraphViz->new(no_overlap => 1, width => 1280, height => 1024);

#	$G = GraphViz->new(width => 640, height => 480, pagewidth => 8.5, 
#		pageheight => 11);

	$G = GraphViz->new(no_overlap => 1, directed => 1, layout => 'dot',
		concentrate => 1);
	if(!$opt_f) {
		print STDERR "\nUsing default log file $DEFAULT_LOG_FILE\n";
		$opt_f = $DEFAULT_LOG_FILE;
	}

	if(!$opt_o) {
		print STDERR "\nUsing default output file $DEFAULT_OUT_FILE\n";
		$opt_o = $DEFAULT_OUT_FILE;
	}

	# open log file
	open(INFILE, $opt_f) or die "Can't open $opt_f: $!";

	# read file line by line
	while(<INFILE>) {
		chomp;

		# check not at the end
		if(/CloseScreen/i) {
			last;
		}

		# check it's something we understand
		if( not /thinc[A-Z].*:/) {
			next;
		}

		# strip off useless text
		# TODO We could keep the time and report it if user wants to
		s/^\(II\)\s//;
		s/^\(EE\)\s//;
		s/^\(WW\)\s//;
        
		s/\((\d){2}:(\d){2}:(\d){2}\)\s//;

		# our state machine. current possibilities:
		# 1. A new object is being created => create a new node in the
		# graph.
		# 2. A Copy is happening => add an edge between the two nodes
		# 3. An object is being destroyed => ???
		# 4. An object is being painted on => ???

		decide: {
			if ($STATE == $ST_INIT) { atInit(); last decide; }
			if ($STATE == $ST_END) { atEnd(); last decide; }
			if ($STATE == $ST_NEWPIX) { atNewPix(); last decide; }
			if ($STATE == $ST_NEWWIN) { atNewWin(); last decide; }
			if ($STATE == $ST_COPY) { atCopy(); last decide; }
			print STDERR "Where am I??? state is $STATE\n";
		}
	}

	close INFILE;
	
	if($STATE != $ST_INIT) {
		print STDERR "ERROR: Finished in an inconsistent state $STATE\n";
	}

	# print stats
	print "\n";
	print "Got $PIX_COUNT pixmaps and $WIN_COUNT windows.\n";
	print "$COPY_COUNT copies were made\n";

	# save the graph. 
	# TODO Add option for format?
	print "\n";
	print "Finished. Do you want to save the graph? ";
	read (STDIN, $ans, 1);
	if($ans eq "y") {
		print "saving it as $opt_o\n";
		$G->as_png($opt_o);
#		$G->as_text('graph.txt');
	} else {
		print "ok, not saving it\n";
	}

	print "\nDone.\n"
}
